/*

"E2xB" COPYRIGHT NOTICE, LICENSE AND DISCLAIMER.

Copyright 2003 by Kostas G. Anagnostakis, Evangelos P. Markatos,
Polyhronakis Mixalhs, Antonatos Spiros


Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, provided 
that the above copyright notice appear in all copies and that both the 
copyright notice and this permission notice and warranty disclaimer appear 
in supporting documentation, and that the names of the authors not be used 
in advertising or publicity pertaining to distribution of the software without 
specific, written prior permission.

The authors disclaim all warranties with regard to this software, including all 
implied warranties of merchantability and fitness.  In no event shall we be liable 
for any special, indirect or consequential damages or any damages whatsoever 
resulting from loss of use, data or profits, whether in an action of contract,
negligence or other tortious action, arising out of or in connection with the
use or performance of this software.

*/


#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <sys/types.h>
#include <string.h>
#include "snort.h"
#include "mstring.h"
#include "mwm.h"
#include "exb.h"


#ifdef EXCLSTATS
int             truematches = 0;
int             falsematches = 0;
int             pktfalse = 0;
int             exbmatches = 0;
int             exbcalls = 0;
int             exbinits = 0;
int             exbsearches = 0;
#endif


static unsigned char xlatcase[256];
static unsigned char xlatcaseu[256];
static char            bits[MAXSIZE];
static char           pktno;
extern PacketCount pc;		/* packet count information */

static INLINE unsigned char *hbm_match(HBM_STRUCT * px, unsigned char *text, int n);
HBM_STRUCT     *hbm_prep(unsigned char *pat, int m);


typedef struct _exb_pat_struct {
	uint8_t      plen;
	uint8_t      nocase;
	HBM_STRUCT  *psBmh;
	int          ID;
	char         *pat;
	struct _exb_pat_struct *next;
}exb_pat_struct;


static INLINE void
ConvCaseToLowerEx(unsigned char *d, unsigned char *s, int m)
{
	int  i;
	for (i = 0; i < m; i++) {
		d[i] = xlatcaseu[s[i]];
	}
}

static INLINE void
ConvCaseToUpperEx(unsigned char *d, unsigned char *s, int m)
{
	int  i;
	
	for (i = 0; i < m; i++) {
		d[i] = xlatcase[s[i]];
	}
}

static void
init_xlatcase()
{
	int i;

	for (i = 0; i < 256; i++) {
		xlatcase[i] = toupper(i);
		xlatcaseu[i] = tolower(i);
	}
}


typedef struct {
	exb_pat_struct *plist;	/* perhaps array would be better */
	exb_pat_struct *ptail;	/* perhaps array would be better */
	int             npat;
	MWM_STRUCT     *mwm;
}exb_struct;

int
exbPrepPatterns(exb_struct * pv)
{
	MWM_STRUCT     *ps = (MWM_STRUCT *) pv->mwm;
	return mwmPrepPatterns(ps);
}
void
exbSetRuleMask(exb_struct * pv, BITOP * rm)
{
	MWM_STRUCT     *ps = (MWM_STRUCT *) pv->mwm;
	ps->RuleMask = rm;
}

void
exbLargeShifts(exb_struct * pv, int flag)
{
	MWM_STRUCT     *ps = (MWM_STRUCT *) pv->mwm;
	ps->msLargeShifts = flag;
}


static INLINE void
init_table(unsigned char *bbuf, const int len)
{
	EREGISTER short  temp;
	EREGISTER unsigned char c1, c2;
	register unsigned char *mmax = bbuf + len - 1;
	c2 = *bbuf;
	while (bbuf++ < mmax) {
		c1 = c2;
		c2 = *bbuf;
		temp = ((c1 << HSIZE) ^ c2);
		*(bits + temp) = pktno;
	}
}

static INLINE uint8_t
mymSearchCI(char *ptrn, const uint8_t plen)
{
	EREGISTER short  temp;
	EREGISTER short  temp2;
	EREGISTER short  temp3;
	EREGISTER short  temp4;
	EREGISTER unsigned char c1, c2, c1l, c2l;

	register unsigned char *c = ptrn;
	register unsigned char *done = ptrn + plen - 1;

	c2=*c;
	c2l=xlatcase[c2];
	while (c < done) {
		c1=*c;
		c2 = *(++c);
		c1l = xlatcase[c1];
		c2l = xlatcase[c2];
		temp = ( c1 << HSIZE) ^ (c2);
		temp2 = ( c1 << HSIZE) ^ (c2l);
		temp3 = ( c1l << HSIZE) ^ (c2);
		temp4 = ( c1l << HSIZE) ^ (c2l);
		if (
		    (*(bits + temp) != pktno) &&
		    (*(bits + temp2) != pktno) &&
		    (*(bits + temp3) != pktno) &&
		    (*(bits + temp4) != pktno))
			return 0;
	}
	return 1;
}

static INLINE int
mymSearch(char *ptrn,  const uint8_t plen)
{


	EREGISTER short  temp;
	register char  *c = ptrn;
	register char  *done = ptrn + plen - 1;

	while (c < done) {
		temp = (((unsigned char) (*c)) << HSIZE) ^ ((unsigned char) *(++c));
		if (*(bits + temp) != pktno) {
			return 0;
		}
	}
	return 1;
}

static char _exbpool[MAXPOOLSZ];
static char *exbpool=NULL;
static size_t exbpoolpos;

void *
exbpmalloc(size_t size)
{
	void *p=exbpool+exbpoolpos;
	exbpoolpos+=size;
	assert(exbpoolpos<MAXPOOLSZ);
	return p;
}


void *exbNew()
{
	exb_struct     *p =NULL;
	if (exbpool==NULL) {
	exbpool=(void *)_exbpool + MAXPOOLSZ/2 - ((uint32_t) _exbpool % (MAXPOOLSZ/2));
	exbpoolpos=0;
	}
	p=(exb_struct *) exbpmalloc(sizeof(exb_struct));
	p->plist = NULL;
	p->npat = 0;
	p->mwm = mwmNew();
	init_xlatcase();
	return p;
}

int
exbAddPatternEx(void *ps, unsigned char *P, int m, unsigned noCase, unsigned offset, unsigned depth, int ID, int IID)
{
	exb_struct     *eps = (exb_struct *) ps;
	MWM_STRUCT     *mwps = (MWM_STRUCT *) eps->mwm;
	char *tmpat=NULL;
	exb_pat_struct *mps = (exb_pat_struct *) exbpmalloc(sizeof(exb_pat_struct));

	eps->npat++;
	
	tmpat = (char *) exbpmalloc((m + 1) * sizeof(char));

	memcpy(tmpat, P, m);
	if (noCase) {
		ConvCaseToLowerEx(tmpat, tmpat, m);
	}
	mps->plen = (uint8_t) m;
	mps->nocase = (uint8_t) noCase;

	mps->ID = ID;
	mps->psBmh = hbm_prep(tmpat, m);
	mps->psBmh->P=tmpat; /* ignore other copy */

	mps->next = eps->plist;
	eps->plist = mps;

	mwmAddPatternEx(mwps, P, m, noCase, offset, depth, ID, IID);
	return 1;
}


INLINE int
exbSearch(const exb_struct * eps,
	  const unsigned char *pkt, const int plen,

	  const int (*match) (unsigned id, int index, void *data),
	  const void *data)
{

	EREGISTER uint8_t             i=0;
	EREGISTER uint8_t             j=0;
	EREGISTER exb_pat_struct *J=(exb_pat_struct *) eps->ptail;
	EREGISTER exb_pat_struct *mps = (exb_pat_struct *) eps->plist;
	static exb_pat_struct *tosearch[MAXTOSEARCH];
	static uint8_t nfound=0;
	register unsigned char *T,*Tx;
#ifdef EXCLSTATS
	int             oldfalse = falsematches;
	exbcalls++;
#endif

	i = pc.total & 0xff;
	if (pktno != i) {
		pktno = i;
#ifdef EXCLSTATS
		exbinits++;
#endif
		init_table(pkt, plen);
	}
	

begin:

	i=0;
	while (mps != NULL) {
#ifdef EXCLSTATS
		exbsearches++;
#endif

#ifndef SHORTSTRUCTS
		if (mps->nocase) {
#else
		if ((mps->plen & 128) >0 ) {
#endif			

		if (mymSearchCI(MPAT(mps), mps->plen) != 0) {
#ifdef EXCLSTATS
				exbmatches++;
#endif
				tosearch[i++] = mps;
				if(i==(MAXTOSEARCH-1)) {
					mps=mps->next;
					goto search;
				}
			
			}
		} else
		
		if (mymSearch(MPAT(mps), mps->plen) != 0) {
#ifdef EXCLSTATS
			exbmatches++;
#endif

			tosearch[i++] = mps;
		
			if(i==(MAXTOSEARCH-1)) {
				mps=mps->next;
				goto search;
			}

		
		}
		
		mps = mps->next;
	}

search:

	T=pkt;
	if (i==0) {
#ifdef EXCLSTATS
		if (falsematches != oldfalse)
			pktfalse++;
#endif
		return (0);
	} else {
		
		
	for (j = 0; j < i; j++) {
		static unsigned char S[65536];
		register uint8_t cflag=0;

		J = tosearch[j];
		T=pkt;
#ifndef SHORTSTRUCTS
		if ((J->nocase)) {
#else
		if ( (J->plen & 128) >0 ) {
#endif			
			T=S;
			if (cflag==0) {
				ConvCaseToLowerEx(S, pkt, plen);
				cflag=1;
			}
		}
		
		if ((Tx=hbm_match(J->psBmh, T, plen)))

		{

#ifdef EXCLSTATS
			truematches++;
#endif
			nfound++;
			if (match(J->ID, (int)(Tx-T), data)) {
				return nfound;
			}
		} 
		else {
		T=pkt;
		
#ifndef SHORTSTRUCTS
			if ((J->nocase)) {
#else
			if ( (J->plen & 128) >0 ) {
#endif
				T=S;
			}
			//printf("false match\n");
#ifdef EXCLSTATS
			falsematches++;
#endif
			}
		}

	goto begin;
	
	}

	return 0; /* never gets here */

}

void
exbStats()
{
#ifdef EXCLSTATS
	printf("calls %d searches %d (%.2f/pkt) matches %d true %d false %d (%.2f %% per pkt %.2f %% per call) inits %d\n", exbcalls, exbsearches, 100.0 * (double) exbsearches / (double) exbcalls, exbmatches, truematches, falsematches, 100.0 * (double) falsematches / (double) exbsearches, 100.0 * (double) pktfalse / (double) exbcalls, exbinits);
#endif
}


static INLINE unsigned char *
hbm_match(HBM_STRUCT * px, unsigned char *text, int n)
{
	unsigned char  *pat, *t, *et, *q;
	int             m1, k;
	short          *bcShift;

	m1 = px->M - 1;
	pat = px->P;
	bcShift = px->bcShift;

	t = text + m1;
	et = text + n;

	/* Handle 1 Byte patterns - it's a faster loop */
	if (!m1) {
		for (; t < et; t++)
			if (*t == *pat)
				return t;
		return 0;
	}
	/* Handle MultiByte Patterns */
	while (t < et) {
		/* Scan Loop - Bad Character Shift */
		do {
			t += bcShift[*t];
			if (t >= et)
				return 0;;

			t += (k = bcShift[*t]);
			if (t >= et)
				return 0;

		} while (k);

		/* Unrolled Match Loop */
		k = m1;
		q = t - m1;
		while (k >= 4) {
			if (pat[k] != q[k])
				goto NoMatch;
			k--;
			if (pat[k] != q[k])
				goto NoMatch;
			k--;
			if (pat[k] != q[k])
				goto NoMatch;
			k--;
			if (pat[k] != q[k])
				goto NoMatch;
			k--;
		}
		/* Finish Match Loop */
		while (k >= 0) {
			if (pat[k] != q[k])
				goto NoMatch;
			k--;
		}
		/* If matched - return 1st char of pattern in text */
		return q;

NoMatch:

		/* Shift by 1, this replaces the good suffix shift */
		t++;
	}

	return 0;
}
